using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using System.Drawing;

namespace CodeJotter {
    public partial class CodeJotter : UserControl {


        protected override void OnKeyPress(KeyPressEventArgs e) {
            InsertTextAtCursor(e.KeyChar.ToString());
            base.OnKeyPress(e);
        }


        protected override bool ProcessCmdKey(ref Message msg, Keys keyData) {


            CursorOn = true;           

            bool ControlKey = (keyData & Keys.Control) == Keys.Control;
            bool ShiftKey = (keyData & Keys.Shift) == Keys.Shift;
            Keys PlainKey = keyData & ~Keys.Modifiers;

            if (ControlKey) {
                switch (PlainKey) {
                    case Keys.V:
                        Paste();
                        return true;
                    case Keys.Up:
                        --TopmostLine;
                        Invalidate();
                        return true;
                    case Keys.Down:
                        ++TopmostLine;
                        Invalidate();
                        return true;
                    case Keys.A:
                        SelectAll();
                        return true;
                    case Keys.C:
                    case Keys.Insert:
                        Copy();
                        return true;
                    case Keys.X:
                        Cut();
                        return true;
                    case Keys.Left:
                        UpdatePointers();
                        if (currentChar.Previous == null) {
                            MoveCursorLeft();
                        } else {
                            while (IsWhitespace(currentChar.Previous.Value)) MoveCursorLeft();
                            while (currentChar != null && currentChar.Previous != null && !IsWhitespace(currentChar.Previous.Value)) {
                                MoveCursorLeft();
                            }
                        }
                        if (!ShiftKey) SelectionCursorPos = CursorPos;
                        Invalidate();
                        return true;
                    case Keys.Right:
                        UpdatePointers();
                        if (currentChar.Next == null) {
                            MoveCursorRight();
                        } else {
                            while (IsWhitespace(currentChar.Value)) MoveCursorRight();
                            while (!IsWhitespace(currentChar.Value)) MoveCursorRight();
                        }
                        if (!ShiftKey) SelectionCursorPos = CursorPos;
                        Invalidate();
                        return true;
                }
                
            }

            if (ShiftKey) {
                switch (PlainKey) {
                    case Keys.Insert:
                        Paste();
                        return true;
                    case Keys.Delete:
                        Cut();
                        return true;
                }
            }

            switch (PlainKey) {
                case Keys.Home:
                    CheckUpdatedPointers();
                    CursorX = 0;
                    if (ControlKey) CursorY = 0;
                    if (!ShiftKey) {
                        SelectionCursorPos = CursorPos;
                    }
                    JumpToCursor();
                    Invalidate();
                    return true;
                case Keys.End:
                    CheckUpdatedPointers();
                    if (ControlKey) CursorY = lines.Count - 1;
                    CursorX = currentLine.Value.Characters.Count - 1;
                    if (!ShiftKey) {
                        SelectionCursorPos = CursorPos;
                    }
                    JumpToCursor();
                    Invalidate();
                    return true;
                case Keys.Left:
                    MoveCursorLeft();
                    if (!ShiftKey) SelectionCursorPos = CursorPos;
                    Invalidate();
                    return true;
                case Keys.Right:
                    MoveCursorRight();
                    if (!ShiftKey) SelectionCursorPos = CursorPos;
                    Invalidate();
                    return true;
                case Keys.Up:
                    --CursorY;
                    if (!ShiftKey) {
                        SelectionCursorPos = CursorPos;
                    }
                    JumpToCursor();
                    Invalidate();
                    return true;
                case Keys.Down:
                    ++CursorY;
                    if (!ShiftKey) {
                        SelectionCursorPos = CursorPos;
                    }
                    JumpToCursor();
                    Invalidate();
                    return true;
                case Keys.PageUp:
                    CursorY -= visibleLines;
                    if (!ShiftKey) {
                        SelectionCursorPos = CursorPos;
                    }
                    JumpToCursor();
                    Invalidate();
                    return true;
                case Keys.PageDown:
                    CursorY += visibleLines;
                    if (!ShiftKey) {
                        SelectionCursorPos = CursorPos;
                    }
                    JumpToCursor();
                    Invalidate();
                    return true;
                case Keys.Delete:
                    if (HasSomethingSelected) {
                        ClearSelection();
                    } else {
                        UpdatePointers();
                        if (currentLine != null && currentChar != null) {
                            if (currentChar.Next != null) {
                                currentLine.Value.Characters.Remove(currentChar);
                            } else {
                                if (cursor.Y != lines.Count - 1) {
                                    foreach (char c in currentLine.Next.Value.Characters) {
                                        currentLine.Value.Characters.AddBefore(currentChar, c);
                                    }
                                    lines.Remove(currentLine.Next);
                                }
                            }
                            UpdatePointers();
                            if (currentLine != null) currentLine.Value.NeedsHighlighting = true;
                        }
                    }
                    JumpToCursor();
                    Invalidate();
                    return true;
                case Keys.Back:
                    if (HasSomethingSelected) {
                        ClearSelection();
                        JumpToCursor();
                        Invalidate();
                        return true;
                    }
                    return base.ProcessCmdKey(ref msg, keyData);
                case Keys.Tab:
                    InsertTextAtCursor("\t");
                    JumpToCursor();
                    Invalidate();
                    return true;
                default:
                    return base.ProcessCmdKey(ref msg, keyData);
            }
        }

        private bool IsWhitespace(char c) {
            switch (c) {
                case ' ':
                case '\t':
                case '\r':
                case '\n':
                    return true;
                default:
                    return false;
            }
        }

        private void MoveCursorLeft() {
            CheckUpdatedPointers();
            if (CursorX == 0) {
                if (CursorY == 0) return;
                --CursorY;
                CursorX = currentLine.Value.Characters.Count - 1;
            } else {
                --CursorX;
            }
            JumpToCursor();
        }

        private void MoveCursorRight() {
            CheckUpdatedPointers();
            if (CursorX == currentLine.Value.Characters.Count - 1) {
                if (CursorY == lines.Count - 1) return;
                ++CursorY;
                CursorX = 0;
            } else {
                ++CursorX;
            }
            JumpToCursor();
        }

        bool CtrlPressed = false;
        bool ShiftPressed = false;
        protected override void OnKeyDown(KeyEventArgs e) {
            if (e.KeyCode == Keys.ControlKey) CtrlPressed = true;
            if (e.KeyCode == Keys.ShiftKey) ShiftPressed = true;
            base.OnKeyDown(e);
        }
        protected override void OnKeyUp(KeyEventArgs e) {
            if (e.KeyCode == Keys.ControlKey) CtrlPressed = false;
            if (e.KeyCode == Keys.ShiftKey) ShiftPressed = false;
            base.OnKeyUp(e);
        }


        

    }
}
